#ifdef VIC_ANALOG

#include "Tcl.h"

#include <stdio.h>
#include <stdlib.h>
#include <UMSVideoIO.xh>
#include <UMSStrings.h>

#define UMS_Success(rc) (rc == UMSVideoIO_Success)

static Environment *ev;
static UMSVideoIO *video;

static _IDL_SEQUENCE_UMSVideoIO_RingBufferElement ring_buffer;

static void allocate_ring_buffers(long n_buffers, long buffer_size,
			   _IDL_SEQUENCE_UMSVideoIO_RingBufferElement *ring_buffer);

static void do_analog_yuv_config(long width, long height);

int analog_yuv_init(long width, long height)
{
    UMSVideoIO_ReturnCode rc;

    ev = somGetGlobalEnvironment();

    video = new UMSVideoIO();

    rc = video->open(ev, "/dev/sr1");
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("Cannot open /dev/sr1\n");
#endif
	rc = video->open(ev, "/dev/sr0");
	if (!UMS_Success(rc))
	{
#ifdef DEBUG
	    printf("Cannot open /dev/sr0\n");
#endif
	    delete video;
	    video = NULL;
	    Tcl &t = Tcl::instance();
	    t.evalc("analog_open_failed");
	    return 0;
	}
    }

    rc = video->set_analog_video_format(ev, NTSC);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("set_analog_video_format failed with rc=%d\n", rc);
#endif
    }

    rc = video->set_output_image_format(ev, YUV422);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("set_output_image_format failed with rc=%d\n", rc);
#endif
    }

    long w, h;
    w = 640;
    h = 480;
    rc = video->set_uncompressed_image_size(ev, &w, &h);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("set_uncompressed_image_size failed with rc=%d\n", rc);
#endif
    }

#ifdef DEBUG
    printf("image size is %dx%d\n", w, h);
#endif

    allocate_ring_buffers(4, w * h * 2, &ring_buffer);

    do_analog_yuv_config(width, height);

    return 1;
}

void analog_yuv_close()
{
    UMSVideoIO_ReturnCode rc;
    
    rc = video->close(ev);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("close failed with rc=%d\n", rc);
#endif
    }
    delete video;
}

void analog_yuv_config(long width, long height)
{
    UMSVideoIO_ReturnCode rc;
    
    if (video == NULL)
	return;

    rc = video->close(ev);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("close failed with rc=%d\n", rc);
#endif
    }
    delete video;

    analog_yuv_init(width, height);
}

static void do_analog_yuv_config(long width, long height)
{
    UMSVideoIO_ReturnCode rc;
    
    rc = video->flush_uncompressed_video_out_buffers(ev);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("flush_uncompressed_video_out_buffers failed with rc=%d\n", rc);
#endif
    }

    rc = video->setup_uncompressed_video_out_buffers(ev, &ring_buffer);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("setup_uncompressed_video_out_buffers failed with rc=%d\n", rc);
#endif
    }
}

void analog_yuv_out(const u_char *yp_i, const u_char *up_i, const u_char *vp_i,
		    long width, long height)
{
    UMSVideoIO_ReturnCode rc;
    long index;
    UMSVideoIO_RingBufferElement *buf;
    int rpt, r, s;
    int hskip, wskip;
    long rwidth, rheight;

    /*
    printf("yuv out %dx%d\n", width, height);
    */

    if (video == NULL)
	return;

    rc = video->get_next_uncompressed_video_out_buffer(ev, &index, 1L);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("get_next_uncompressed_video_out_buffer failed rc=%d\n", rc);
#endif
	return;
    }

    buf = &(ring_buffer._buffer[index]);
    buf->InUseByCaller = 1;

    rwidth = width;
    rheight = height;

    wskip = hskip = 0;
    if (width == 160 && height == 120)
    {
	rpt = 4;
    }
    else if (width == 320 && height == 240)
    {
	rpt = 2;
    }
    else if (width == 640 && height == 480)
    {
	rpt = 1;
    }
    else if (width == 352 && height == 288)
    {
	rpt = 2;
	wskip = (352 * 2 - 640) / 2;
	hskip = (288 * 2 - 480) / 2;
	width = 320;
	height = 240;
    }
    else
    {
	return;
    }
    
    long w, h;
    u_char *data, *dp, *yps, *ups, *vps, *yp, *up, *vp;

    yp = (u_char *) yp_i;
    up = (u_char *) up_i;
    vp = (u_char *) vp_i;

    data = (u_char *) buf->Address;
    dp = data;

    h = height;

    yp += hskip * rwidth;
    up += hskip * rwidth / 2;
    vp += hskip * rwidth / 2;
    while (--h >= 0)
    {
	yps = yp;
	ups = up;
	vps = vp;

	for (r = 0; r < rpt; r++)
	{
	    yp = yps + wskip;
	    up = ups + wskip / 2;
	    vp = vps + wskip / 2;
	    
	    for (w = width; w > 0; w -= 2)
	    {
		for (s = 0; s < rpt; s++)
		{
		    dp[0] = yp[0];
		    dp[1] = up[0];
		    dp[2] = yp[1];
		    dp[3] = vp[0];
		    dp += 4;
		}

		yp += 2;
		up++;
		vp++;
	    }
	}
    }

    buf->SizeOfDataInBuffer = 640 * 480 * 2;
    rc = video->put_uncompressed_frame(ev, index);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("put_uncompressed_frame failed rc=%d\n", rc);
#endif
    }
    buf->InUseByCaller = 0;
}

static void allocate_ring_buffers(long n_buffers, long buffer_size,
			   _IDL_SEQUENCE_UMSVideoIO_RingBufferElement *ring_buffer)
{
    unsigned char **Address;
    UMSVideoIO_RingBufferElement *rbuffer1;
    int i;
    
    Address = new unsigned char * [n_buffers];
    
    ring_buffer->_length = n_buffers;
    ring_buffer->_maximum  = n_buffers;

#ifdef DEBUG
    printf("Allocating ring buffers: n_buffers=%d buffer_size=%d\n",
	   n_buffers, buffer_size);
#endif /* DEBUG */

    ring_buffer->_buffer = new UMSVideoIO_RingBufferElement[n_buffers];

    if (ring_buffer->_buffer <= 0)
    {
	fprintf(stderr, "Cannot malloc somThis->ring_buffer->_buffer\n");
	exit(1);
    }

    /* Set up each ring buffer */

    rbuffer1 = ring_buffer->_buffer;

    for (i = 0; i < n_buffers; i++)
    {
	Address[i] = (unsigned char *)
	    malloc(buffer_size + 4096);
	if (Address[i] <= 0)
	{
	    fprintf(stderr, "Cannot malloc ring_buffer %d\n", i);
	    free(ring_buffer->_buffer);
	    exit(1);
	}
	rbuffer1->Address = ( ((long) Address[i]) + 4096
					- ( ((long) Address[i]) % 4096));
	rbuffer1->AfterHeader = rbuffer1->Address;
	rbuffer1->SizeOfBuffer = buffer_size;
	rbuffer1->SizeOfDataInBuffer = 0;
	rbuffer1->InUseByCaller	= 0;
	rbuffer1->NumberOfOverruns = 0;
	rbuffer1++;
    }
}
#endif
