#ifdef VIC_ANALOG

#include <stdio.h>
#include <stdlib.h>
#include <UMSVideoIO.xh>
#include <UMSStrings.h>
#include "Tcl.h"

#define UMS_Success(rc) (rc == UMSVideoIO_Success)

static Environment *ev;
static UMSVideoIO *video;

static _IDL_SEQUENCE_UMSVideoIO_RingBufferElement ring_buffer;
static _IDL_SEQUENCE_octet jpeg_header;
static long jpeg_header_size;

static void allocate_ring_buffers(long n_buffers, long buffer_size,
			   _IDL_SEQUENCE_UMSVideoIO_RingBufferElement *ring_buffer);
static void get_jpeg_header(long width, long height, long q);

static void do_analog_config(long width, long height, long q);

int analog_init(long width, long height, long q)
{
    UMSVideoIO_ReturnCode rc;

    ev = somGetGlobalEnvironment();

    video = new UMSVideoIO();

    rc = video->open(ev, "/dev/sr1");
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("Cannot open /dev/sr1\n");
#endif
	rc = video->open(ev, "/dev/sr0");
	if (!UMS_Success(rc))
	{
#ifdef DEBUG
	    printf("Cannot open /dev/sr0\n");
#endif
	    delete video;
	    video = NULL;
	    Tcl &t = Tcl::instance();
	    t.evalc("analog_open_failed");
	    return 0;
	}
    }

    rc = video->set_analog_video_format(ev, NTSC);
#ifdef DEBUG
    if (!UMS_Success(rc))
	printf("set_analog_video_format failed with rc=%d\n", rc);
#endif

    allocate_ring_buffers(4, 100000, &ring_buffer);

    do_analog_config(width, height, q);
    return 1;
}

void analog_close()
{
    UMSVideoIO_ReturnCode rc;
    
    rc = video->close(ev);
#ifdef DEBUG
    if (!UMS_Success(rc))
	printf("close failed with rc=%d\n", rc);
#endif
    delete video;
}

static void get_jpeg_header(long width, long height, long q)
{
    UMSVideoIO_ReturnCode rc;
    
#ifdef DEBUG
    printf("Setting image to %dx%d q=%d\n", width, height, q);
#endif
    rc = video->set_input_image_size(ev, &width, &height);
    if (!UMS_Success(rc) && rc != UMSVideoIO_ValueChanged)
    {
#ifdef DEBUG
	printf("set_input_image_size %dx%d failed with rc=%d\n",
	       width, height, rc);
#endif
    }
    rc = video->set_quality_factor(ev, &q);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("set_quality_factor failed with rc=%d\n", rc);
#endif
	return;
    }
#ifdef DEBUG
    printf("Set image to %dx%d q=%d\n", width, height, q);
#endif
    rc = video->get_jpeg_header_size(ev, &jpeg_header_size,
				     UMSVideoIO_WithDQTandDHT);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("get_jpeg_header_size failed with rc=%d\n", rc);
#endif
	return;
    }

#ifdef DEBUG
    printf("Got header size %d\n", jpeg_header_size);
#endif
    
    jpeg_header._buffer = new octet [jpeg_header_size];
    jpeg_header._maximum = jpeg_header._length = jpeg_header_size;

    rc = video->get_jpeg_header(ev, &jpeg_header, UMSVideoIO_WithDQTandDHT);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("get_jpeg_header failed with rc=%d\n", rc);
#endif
	return;
    }
}

void analog_config(long width, long height, long q)
{
    UMSVideoIO_ReturnCode rc;
    
    if (video == NULL)
	return;

    rc = video->close(ev);
#ifdef DEBUG
    if (!UMS_Success(rc))
	printf("close failed with rc=%d\n", rc);
#endif
    delete video;

    analog_init(width, height, q);
}

static void do_analog_config(long width, long height, long q)
{
    UMSVideoIO_ReturnCode rc;
    
    rc = video->flush_compressed_video_out_buffers(ev);
#ifdef DEBUG
    if (!UMS_Success(rc))
	printf("flush_compressed_video_out_buffers failed with rc=%d\n", rc);
#endif

    get_jpeg_header(width, height, q);

    rc = video->setup_compressed_video_out_buffers(ev, &ring_buffer);
#ifdef DEBUG
    if (!UMS_Success(rc))
	printf("setup_compressed_video_out_buffers failed with rc=%d\n", rc);
#endif

    rc = video->put_jpeg_header(ev, &jpeg_header);
#ifdef DEBUG
    if (!UMS_Success(rc))
	printf("put_jpeg_header failed with rc=%d\n", rc);
#endif

}

void analog_out(u_char *frame, int len)
{
    UMSVideoIO_ReturnCode rc;
    long index;
    UMSVideoIO_RingBufferElement *buf;
    
    if (video == NULL)
	return;

    rc = video->get_next_compressed_video_out_buffer(ev, &index, 1L);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("get_next_compressed_video_out_buffer failed rc=%d\n", rc);
#endif
	return;
    }

    buf = &(ring_buffer._buffer[index]);
    buf->InUseByCaller = 1;
    bcopy((char *) frame, (char *) buf->Address, len);
    buf->SizeOfDataInBuffer = len;
    rc = video->put_compressed_frame(ev, index);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("put_compressed_frame failed rc=%d\n", rc);
#endif
    }
    buf->InUseByCaller = 0;
}

static void allocate_ring_buffers(long n_buffers, long buffer_size,
			   _IDL_SEQUENCE_UMSVideoIO_RingBufferElement *ring_buffer)
{
    unsigned char **Address;
    UMSVideoIO_RingBufferElement *rbuffer1;
    int i;
    
    Address = new unsigned char * [n_buffers];
    
    ring_buffer->_length = n_buffers;
    ring_buffer->_maximum  = n_buffers;

#ifdef DEBUG
    printf("Allocating ring buffers: n_buffers=%d buffer_size=%d\n",
	   n_buffers, buffer_size);
#endif /* DEBUG */

    ring_buffer->_buffer = new UMSVideoIO_RingBufferElement[n_buffers];

    if (ring_buffer->_buffer <= 0)
    {
	fprintf(stderr, "Cannot malloc somThis->ring_buffer->_buffer\n");
	exit(1);
    }

    /* Set up each ring buffer */

    rbuffer1 = ring_buffer->_buffer;

    for (i = 0; i < n_buffers; i++)
    {
	Address[i] = (unsigned char *)
	    malloc(buffer_size + 4096);
	if (Address[i] <= 0)
	{
	    fprintf(stderr, "Cannot malloc ring_buffer %d\n", i);
	    free(ring_buffer->_buffer);
	    exit(1);
	}
	rbuffer1->Address = ( ((long) Address[i]) + 4096
					- ( ((long) Address[i]) % 4096));
	rbuffer1->AfterHeader = rbuffer1->Address;
	rbuffer1->SizeOfBuffer = buffer_size;
	rbuffer1->SizeOfDataInBuffer = 0;
	rbuffer1->InUseByCaller	= 0;
	rbuffer1->NumberOfOverruns = 0;
	rbuffer1++;
    }
}
#endif
