/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/color.C,v 1.3 1995/08/26 21:15:40 olson Exp (LBL)";
#endif

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <osfcn.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <string.h>
#include "color.h"
#include "color-models.h"
#include "Tcl.h"

#ifdef _AIX
extern "C" int strcasecmp (const char *, const char *);
#endif

#define COLORMODEL_TRUE		0
#define COLORMODEL_DITHER	1
#define COLORMODEL_QUANTIZE	2
#define COLORMODEL_MONO		3
#define COLORMODEL_GRAY		4
#define COLORMODEL_TRUEGRAY	5
#define COLORMODEL_JV		6
#define COLORMODEL_BEST		7
#define COLORMODEL_ED		8
struct cment {
	const char* name;
	int model;
} cments[] = {
	{ "true",	COLORMODEL_TRUE },
	{ "dither",	COLORMODEL_DITHER },
	{ "od",		COLORMODEL_DITHER },
	{ "quantize",	COLORMODEL_QUANTIZE },
	{ "quant",	COLORMODEL_QUANTIZE },
	{ "mono",	COLORMODEL_MONO },
	{ "gray",	COLORMODEL_GRAY },
	{ "grey",	COLORMODEL_GRAY },
	{ "truegray",	COLORMODEL_TRUEGRAY },
	{ "truegrey",	COLORMODEL_TRUEGRAY },
	{ "jv",		COLORMODEL_JV },
	{ "ed",		COLORMODEL_ED },
	{ "best",	COLORMODEL_BEST },
	{ 0,		-1 },
};

int ColorModel::atocm(const char* s)
{
	for (cment* p = cments; p->name != 0; ++p)
		if (strcasecmp(p->name, s) == 0)
			return (p->model);

	fprintf(stderr, "vic: unknown color model `%s'\n", s);
	exit(1);
}

const char* ColorModel::cmtoa(int model)
{
	for (cment* p = cments; p->name != 0; ++p)
		if (p->model == model)
			return (p->name);

	return ("<illegal>");
}

void
cmap_full()
{
	fprintf(stderr, "vic: private colormap full\n");
	exit(1);
}

void ColorModel::init(int colormodel)
{
	switch (colormodel) {

	case COLORMODEL_BEST:
		instance_ = new TrueColorModel();
		if (!instance_->valid()) {
			delete instance_;
			instance_ = new EDColorModel();
			if (!instance_->valid()) {
				delete instance_;
				instance_ = new QuantColorModel();
			}
		}
		break;

	case COLORMODEL_TRUE:
		instance_ = new TrueColorModel();
		break;

	case COLORMODEL_DITHER:
		instance_ = new DitherColorModel();
		break;

	case COLORMODEL_ED:
		instance_ = new EDColorModel();
		break;

	case COLORMODEL_QUANTIZE:
		instance_ = new QuantColorModel();
		break;

	case COLORMODEL_GRAY:
		instance_ = new GrayColorModel();
		break;

	case COLORMODEL_TRUEGRAY:
		instance_ = new TrueGrayColorModel();
		break;

	case COLORMODEL_JV:
		instance_ = new DitherColorModel(1);
		break;

	case COLORMODEL_MONO:
		instance_ = new MonoColorModel();
		break;
	}
	if (instance_ == 0 || !instance_->valid()) {
		fprintf(stderr,
			"vic: can't use %s color model on this display\n",
			cmtoa(colormodel));
		exit(1);
	}
	if (use_private_map_)
		instance_->private_map();

	if (instance_->alloc_grays() < 0 ||
	    instance_->alloc_colors() < 0) {
		if (use_private_map_)
			cmap_full();
		fprintf(stderr, "vic: can't allocate color map\n");
		instance_->free_colors();
		instance_->private_map();
		if (instance_->alloc_grays() < 0 ||
		    instance_->alloc_colors() < 0)
			cmap_full();
		fprintf(stderr, "vic: reverting to private color map\n");
		use_private_map_ = 1;
	}
}

void ColorModel::release()
{
	delete instance_;
	instance_ = 0;
}

ColorModel* ColorModel::instance_;
int ColorModel::use_private_map_;

ColorModel::ColorModel() : ncolor_(0), valid_(0), eflag_(0)
{
	Tcl& tcl = Tcl::instance();
	dpy_ = Tk_Display(tcl.tkmain());
	/* set up defaults */
	int screen = DefaultScreen(dpy_);
	depth_ = DefaultDepth(dpy_, screen);
	visual_ = DefaultVisual(dpy_, screen);
	colormap_ = XDefaultColormap(dpy_, screen);
	memset((char*)colors_, 0, sizeof(colors_));
	const char* s = tcl.attr("gamma");
	gamma_ = (s != 0) ? atof(s) : 1.;
	if (gamma_ < 0) {
		fprintf(stderr, "vic: bad gamma value %g (ignoring)\n",
			gamma_);
		gamma_ = 1.;
	}
}

ColorModel::~ColorModel()
{
	free_colors();
}

int ColorModel::newLUT(const char* file)
{
	return (-1);
}

void ColorModel::rgb_to_yuv(color& c) const
{
	double r = c.r;
	double g = c.g;
	double b = c.b;

	c.y = short(0.299 * r + 0.587 * g + 0.114 * b + 0.5);
	c.u = short(-0.1687 * r - 0.3313 * g + 0.5 * b + 0.5) + 128;
	c.v = short(0.5 * r - 0.4187 * g - 0.0813 * b + 0.5) + 128;
}

#ifdef notdef
int ColorModel::yuv_to_rgb(color& c) const
{
	double y = c.y;
	double u = c.u - 128.;
	double v = c.v - 128.;

	int r = int(y + 1.402 * v + 0.5);
	int g = int(y - 0.34414 * u - 0.71414 * v + 0.5);
	int b = int(y + 1.772 * u + 0.5);

	int valid = 1;
#define CHECK(v) \
	if (v > 255) { \
		v = 255; \
		valid = 0; \
	} else if (v < 0) { \
		v = 0; \
		valid = 0; \
	}
	CHECK(r)
	c.r = r;
	CHECK(g)
	c.g = g;
	CHECK(b)
	c.b = b;

	return (valid);
}
#endif

/*
 * Allocate a private colormap and attach it to the top-level window.
 * Allocate the first 32 cells using the same colors as the default
 * colormap in order to minimize technicolor (i.e., since servers
 * allocate entries in order and colormap fragmentation isn't all
 * that common, most of the important colors, e.g., those used in
 * the user's default configuration, will be retained).
 * XXX should use smarter approach that uses color distances to match
 * colormaps up as best as possible.
 */
void ColorModel::private_map()
{
	Colormap dm = colormap_;
	Tk_Window tk = Tcl::instance().tkmain();
	colormap_ = XCreateColormap(dpy_, Tk_WindowId(tk), visual_, AllocNone);
	Tk_SetWindowColormap(tk, colormap_);
	XColor colors[32];
	int i;
	for (i = 0; i < 32; ++i)
		colors[i].pixel = i;
	XQueryColors(dpy_, dm, colors, 32);
	for (i = 0; i < 32; ++i) {
		color c;
		c.r = colors[i].red >> 8;
		c.g = colors[i].green >> 8;
		c.b = colors[i].blue >> 8;
		alloc_color(c);
	}
}

void ColorModel::free_colors()
{
	if (ncolor_ > 0) {
		XFreeColors(dpy_, colormap_, pixel_, ncolor_, 0);
		ncolor_ = 0;
	}
}

int ColorModel::alloc_color(color& c)
{
	int r = c.r;
	int g = c.g;
	int b = c.b;
	if (gamma_ != 1.) {
		r = int(256. * pow(r / 256., gamma_));
		g = int(256. * pow(g / 256., gamma_));
		b = int(256. * pow(b / 256., gamma_));
	}
	if (r == g && r == b && (r & 7) == 0 && r > 0) {
		/*
		 * This color is one level away from the gray ramp
		 * used in nv, ghostscript, and the LBL-modified
		 * tk library.  Change it so we'll end up sharing
		 * the colormap entry.  The error won't be perceptible.
		 */
		 r -= 1;
		 g -= 1;
		 b -= 1;
	 }

	XColor xc;
	xc.red = r << 8;
	xc.green = g << 8;
	xc.blue = b << 8;
	if (XAllocColor(dpy_, colormap_, &xc) == 0) {
		free_colors();
		return (-1);
	}
	int pixel = xc.pixel;
	color& p = colors_[pixel];
	pixel_[ncolor_] = pixel;
	if (++ncolor_ >= 256) {
		fprintf(stderr, "vic: colormap overflow (internal error)\n");
		exit(1);
	}
	p.r = xc.red >> 8;
	p.g = xc.green >> 8;
	p.b = xc.blue >> 8;

	/* XXX ed ditherer needs this */
	rgb_to_yuv(p);

	return (pixel);
}

int ColorModel::alloc_colors()
{
	return (0);
}

int ColorModel::alloc_grays()
{
#define DELTA (256 / 32)
	int pixel[32];
	int i;
	for (i = 0; i < 32; ++i) {
		int luma = (i + 1) * DELTA - 1;
		color c;
		c.r = c.g = c.b = luma;
		pixel[i] = alloc_color(c);
		if (pixel[i] < 0)
			return (-1);
	}
	for (i = 0; i < 256; ++i) {
		int level = (i + 1) / DELTA;
		int err;
		if (level == 0)
			err = 0;
		else {
			level -= 1;
			err = (i + 1) & (DELTA - 1);
		}
		grey_lut_[i] = pixel[level] | err << 8;
	}
	for (; i < 512; ++i)
		grey_lut_[i] = pixel[31];

	return (0);
}

Colorizer::Colorizer(int inw, int inh, int scale)
	: inw_(inw), inh_(inh), scale_(scale)
{
}

void ColorModel::fork(const u_int* histogram)
{
	if (proc_.dump(histogram) >= 0)
		proc_.fork(eflag_);
}

MedianCutHandler::MedianCutHandler() : child_(0)
{
	int pid = getpid();
	sprintf(infile_, "/tmp/vicHIST.%u", pid);
	sprintf(outfile_, "/tmp/vicLUT.%u", pid);
}

MedianCutHandler::~MedianCutHandler()
{
	(void)::unlink(infile_);
	(void)::unlink(outfile_);
}

#define FAILED_EXEC 123
void MedianCutHandler::dispatch(int mask)
{
	int status;
	waitpid(child_, &status, 0);
	status = WEXITSTATUS(status);
	child_ = 0;
	(void)::unlink(infile_);
	if (status == 0)
		(void)ColorModel::instance().newLUT(outfile_);
	else if (status == FAILED_EXEC)
		fprintf(stderr, "vic: couldn't exec histtolut\n");
	else
		fprintf(stderr, "vic: histtolut exited with status %d\n",
			status);

	(void)::unlink(outfile_);
	unlink();
	close(channel_);
	Tcl::instance().evalc("finish_mediancut");
}

int MedianCutHandler::dump(const u_int* histogram)
{
	int fd = open(infile_, O_WRONLY|O_TRUNC|O_CREAT, 0644);
	if (fd < 0) {
		/*XXX*/
		perror(infile_);
		return (-1);
	}
	write(fd, (char*)histogram, 65536 * sizeof(*histogram));
	close(fd);
	return (0);
}

extern "C" pid_t vfork();

int MedianCutHandler::fork(int eflag)
{
	char* av[8];
	char* program = "histtolut"; /*XXX*/
	av[0] = program;
	av[1] = "-n";
	av[2] = (char*)Tcl::instance().attr("medianCutColors");
	av[3] = "-o";
	av[4] = outfile_;
	if (eflag) {
		av[5] = "-e";
		av[6] = infile_;
		av[7] = 0;
	} else {
		av[5] = infile_;
		av[6] = 0;
	}
	/*
	 * A pipe for communicating between the parent and child.
	 * The naming convention is from the perspective of the child.
	 * We arrange for the child's stdout to be directed at the parent.
	 */
	int pout[2];
	if (pipe(pout) < 0) {
		perror("pipe");
		exit(1);
	}
	/* SGI doesn't have vfork */
#if !defined(sgi) && !defined(_AIX)
	int pid = ::vfork();
#else
	int pid = ::fork();
#endif
	if (pid < 0) {
		perror("vfork");
		exit(1);
	}
	if (pid == 0) {
		/*
		 * The child -- exec the new process.
		 * Bind pipe descriptor to stdio.
		 */
		close(pout[0]);
		dup2(pout[1], 1);

		for (int i = 3; i < 32; ++i)
			close(i);

		execvp(program, av);
		_exit(FAILED_EXEC);
	}
	/*
	 * The parent -- set up non-blocking I/O and
	 * remember our side of the pipe.
	 */
	close(pout[1]);
	child_ = pid;
	channel_ = pout[0];
	link(channel_, TK_READABLE);

	return (0);
}
