/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef vic_decoder_h
#define vic_decoder_h

extern "C" {
#include <tk.h>
}

#include <sys/param.h>
#include <sys/uio.h>

#ifdef _AIX
extern "C" int strcasecmp (const char *, const char *);
#endif

#include "inet.h"
#include "packet.h"
#include "iohandler.h"
#include "timer.h"
#include "video.h"
#include "source.h"

class VideoWindow;
class SoftwareRenderer;
class SoftwareDecoder;

struct dmabuf {
	int shmid;
	u_char* bp;
};

class Decoder : public PacketHandler, public Timer {
 protected:
	Decoder(int hdrlen);
	virtual void update_stamp();
 public:
	virtual ~Decoder();
	inline int width() const { return (inw_); }
	inline int height() const { return (inh_); }
	virtual void info(char*) const;
	virtual void counter_names(char*) const;
	virtual int counter_query(const char* name, u_int& value) const;

	static Decoder* create(int fmt, const u_char* pkthdr);
	virtual int recv(rtphdr*, u_char* vh, u_char* data,
			 int len, int jitter);
	virtual void scale_rate(int severity);
	virtual void timeout();
	virtual int colorhist(u_int* histogram) const;

	/*
	 * Add or remove a window to the set of windows
	 * that this decoder outputs to.
	 */
	virtual void attach(VideoWindow*, int stamp);
	virtual void detach(VideoWindow*);

	inline int valid() { return (valid_); }
	inline int color() { return (color_); }

	static void allocshm(dmabuf&, int, int flag = 0);
	static void freeshm(dmabuf&);
	virtual void setcolor(int gray);
	static double gettimeofday();

	static inline void use_hardware(int v) { use_hardware_ = v; }
	virtual void redraw();

	virtual void toggle_analog();
 protected:
	static int use_hardware_;

	int analog_;

	int valid_;
	int color_;
	int inw_;		/* native image width */
	int inh_;		/* native image height */
	int ntoss_;		/* count of late packets */

	int update_;		/* update interval for thumbnails */
	int update_stamp_;	/* true when thumbnail needs to be updated */
};

/*
 * Rendering vector.  We keep a vector of timestamps of when each individual
 * block of frame has been updated by a decoder.  Then, different renderers
 * can keep track of the last time they've updated their view of the frame
 * (and dither only those blocks that need to be updated).
 *
 * We size the data structure according to the largest it can possibly
 * be (i.e., full PAL).  No big deal since it's block-oriented and
 * therefore small.  This avoids reallocation on size changes and by
 * making the width of the array a power of two, gives efficient indexing.
 */
#define RV_MAXWIDTH 1024
#define RV_INDEX(x, y) ((y) << 7 | (x))
#define RV_MAXBLKS (RV_MAXWIDTH * 576 / 64)
#define RV_STRIDE (RV_MAXWIDTH / 8)

class SoftwareDecoder : public Decoder {
 public:
	SoftwareDecoder(int hdrlen);
	virtual ~SoftwareDecoder();
	virtual void attach(VideoWindow*, int stamp);
	virtual void detach(VideoWindow*);
	virtual void setcolor(int gray);
	virtual int colorhist(u_int* histogram) const = 0;
	virtual void redraw();
 protected:
	virtual SoftwareRenderer* create(VideoWindow*) = 0;
	virtual void redraw(SoftwareRenderer*) = 0;
	SoftwareRenderer* stamp_;
	SoftwareRenderer* engines_;

	void render_frame(const u_char* yp, const u_char* up,
			  const u_char* vp);
	int now_;
	u_char rvts_[RV_MAXBLKS];

	void colorhist_411_556(u_int* hist, const u_char* y, const u_char* u,
			       const u_char* v, int width, int height) const;
	void colorhist_422_556(u_int* hist, const u_char* y, const u_char* u,
			       const u_char* v, int width, int height) const;
};

class PlaneDecoder : public SoftwareDecoder {
 public:
	PlaneDecoder(int hdrlen);
	virtual ~PlaneDecoder();
	virtual int colorhist(u_int* histogram) const;
 protected:
	void resize(int width, int height);
	virtual void redraw(SoftwareRenderer *p);
	SoftwareRenderer* create(VideoWindow* w);

	u_char* yplane_;
	u_char* uplane_;
	u_char* vplane_;
};

#endif
