/*
 * Copyright (c) 1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/encoder-h261.C,v 1.3 1995/08/26 21:15:49 olson Exp (LBL)";

#include <stdio.h>
#include <string.h>
#include <osfcn.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/time.h>
#include "inet.h"
#include "net.h"
#include "packet.h"
#include "encoder-h261.h"
#include "dct.h"
#include "p64/p64-huff.h"
#include "endian.h"
#include "Tcl.h"
#include "crdef.h"

#define HDRSIZE (sizeof(rtphdr) + 2)

H261Encoder::H261Encoder(Network* net, int mtu) : YuvEncoder(net, mtu),
	  bs_(0), es_(0), ngob_(12)
{
	format_ = RTP_PT_H261;
	setq(10);

	int size = 8*1024;
	buffer_ = new u_char[size];
	es_ = &buffer_[size];
}

H261Encoder::~H261Encoder()
{
	delete buffer_;
}

void H261Encoder::grow()
{
	int off = bs_ - buffer_;
	int size = 2 * (es_ - buffer_);
	u_char* p = new u_char[size];
	memcpy(p, buffer_, size >> 1);
	delete buffer_;
	buffer_ = p;
	bs_ = p + off;
	es_ = p + size;
}

/*
 * Set up the forward DCT quantization table for
 * INTRA mode operation.
 */
void H261Encoder::setq(int q)
{
	quant_ = q;
	mquant_ = q;
	/*
	 * Set the DC quantizer to 1, since we want to do this
	 * coefficient differently (i.e., the DC is rounded while
	 * the AC terms are truncated).
	 */
	qt_[0] = 1;
	q <<= 1;
	for (int i = 1; i < 64; ++i)
		qt_[i] = q;

	fdct_fold_q(qt_, qt_, qinv_);
}

#define MBPERGOB 33

void H261Encoder::size(int w, int h)
{
	if (w == 352 && h == 288) {
		cif_ = 1;
		ngob_ = 12;
		bstride_ = 22 + 44;
		lstride_ = 16 * 352 - 352 / 2;
		cstride_ = 8 * 176 - 176 / 2;
	} else if (w == 176 && h == 144) {
		cif_ = 0;
		ngob_ = 6; /* not really number of GOBs, just loop limit */
		bstride_ = 22;
		lstride_ = 16 * 176 - 176;
		cstride_ = 8 * 88 - 88;
	} else {
		/*XXX*/
		fprintf(stderr, "vic: H.261 bad geometry: %dx%d\n", w, h);
		exit(1);
	}
	int loff = 0;
	int coff = 0;
	int blkno = 0;
	for (int gob = 0; gob < ngob_; gob += 2) {
		loff_[gob] = loff;
		coff_[gob] = coff;
		blkno_[gob] = blkno;
		/* width of a GOB (these aren't ref'd in QCIF case) */
		loff_[gob + 1] = loff + 11 * 16;
		coff_[gob + 1] = coff + 11 * 8;
		blkno_[gob + 1] = blkno + 11 * 2;

		/* advance to next GOB row */
		loff += (16 * 16 * MBPERGOB) << cif_;
		coff += (8 * 8 * MBPERGOB) << cif_;
		blkno += (4 * MBPERGOB) << cif_;
	}
	framesize_ = w * h;
	width_ = w;
	height_ = h;
}

int H261Encoder::command(Tcl& t, int argc, char** argv)
{
	if (argc == 3 && strcmp(argv[1], "q") == 0) {
		setq(atoi(argv[2]));
		return (TCL_OK);
	}
	return (Framer::command(t, argc, argv));
}

/*XXX use 64-bits if available */
#define NBIT 32

#if BYTE_ORDER == LITTLE_ENDIAN
#define STORE_BITS(bs, bb) \
	((u_char*)bs)[0] = bb >> 24; \
	((u_char*)bs)[1] = bb >> 16; \
	((u_char*)bs)[2] = bb >> 8; \
	((u_char*)bs)[3] = bb;
#else
#define STORE_BITS(bs, bb) *(u_int*)(bs) = (bb);
#endif

#define PUT_BITS(bits, n, nbb, bb, bs) \
{ \
	nbb += (n); \
	if (nbb > NBIT)  { \
		int extra = (nbb) - NBIT; \
		bb |= (bits) >> extra; \
		STORE_BITS(bs, bb) \
		bs += (NBIT / 8); \
		bb = (bits) << NBIT - extra; \
		nbb = extra; \
	} else \
		bb |= (bits) << NBIT - (nbb); \
}

void H261Encoder::encode_blk(short* blk)
{
	u_int bb = bb_;
	u_int nbb = nbb_;
	/*
	 * Quantize DC.  Round instead of truncate.
	 */
	int dc = (blk[0] + 4) >> 3;
	if (dc <= 0)
		/* shouldn't happen with CCIR 601 black (level 16) */
		dc = 1;
	else if (dc > 254)
		dc = 254;
	else if (dc == 128)
		/* per Table 6/H.261 */
		dc = 255;
	/* Code DC */
	PUT_BITS(dc, 8, nbb, bb, bs_);
	int run = 0;
	for (int i = 1; i < 64; ++i) {
		if (bs_ >= es_) {
			fprintf(stderr,
				"H261Encoder::encode_blk: buffer overflow\n");
			exit(1);
		}
		int k = COLZAG[i];
#ifdef H261_ENCODE_UPPER_QUAD_ONLY
		if ((k & 0x24) != 0) {
			++run;
			continue;
		}
#endif
		int level = blk[k];
		if (level == 0) {
			++run;
			continue;
		}
		if (run <= 26 && level <= 15 && level >= -15) {
			/*
			 * See if we can use a VLC.
			 */
			huffent* he = &hte_tc[(level & 0x1f) << 5 | run];
			if (he->nb != 0) {
				PUT_BITS(he->val, he->nb, nbb, bb, bs_);
				run = 0;
				continue;
			}
		}
		/*
		 * Can't use a VLC.  Escape it.
		 */
		int code = 1 << 14;
		code |= run << 8;
		code |= level & 0xff;
		PUT_BITS(code, 20, nbb, bb, bs_);
		run = 0;
	}
	/* EOB */
	PUT_BITS(2, 2, nbb, bb, bs_);

	bb_ = bb;
	nbb_ = nbb;
}

#define MAXMBSIZE (16*16+2*8*8)

void H261Encoder::encode_mb(int mba, const u_char* frm, int loff, int coff)
{
	if (es_ - bs_ < MAXMBSIZE)
		grow();

	/*
	 * encode all 6 blocks of the macro block to find the largest
	 * coef (so we can pick a new quantizer if gquant doesn't have
	 * enough range).
	 */
	/*XXX this can be u_char instead of short but need smarts in fdct */
	short blk[64 * 6];
	register int stride = width_;
	/* luminance */
	const u_char* p = &frm[loff];
	fdct(p, stride, blk + 0, qt_, qinv_);
	fdct(p + 8, stride, blk + 64, qt_, qinv_);
	fdct(p + 8 * stride, stride, blk + 128, qt_, qinv_);
	fdct(p + (8 * stride + 8), stride, blk + 192, qt_, qinv_);
	/* chominance */
	int fs = framesize_;
	p = &frm[fs + coff];
	stride >>= 1;
	fdct(p, stride, blk + 256, qt_, qinv_);
	fdct(p + (fs >> 2), stride, blk + 320, qt_, qinv_);

	/*
	 * if the default quantizer is too small to handle the coef.
	 * dynamic range, spin through the blocks and see if any
	 * coef. would significantly overflow.
	 */
	register int q = quant_;
	if (q < 8) {
		register int cmin = 0, cmax = 0;
		register short* bp = blk;
		for (register int i = 6; --i >= 0; ) {
			++bp;	// ignore dc coef
			for (register int j = 63; --j >= 0; ) {
				register int v = *bp++;
				if (v < cmin)
					cmin = v;
				else if (v > cmax)
					cmax = v;
			}
		}
		if (cmax < -cmin)
			cmax = -cmin;
		if (cmax >= 128) {
			/* need to re-quantize */
		    register int s;
			for (s = 1; cmax >= (128 << s); ++s) {
			}
			q <<= s;
			register short* bp = blk;
			for (register int i = 6; --i >= 0; ) {
				++bp;	// ignore dc coef
				for (register int j = 63; --j >= 0; ) {
					register int v = *bp;
					*bp++ = v >> s;
				}
			}
		}
	}

	int m = mba - mba_;
	mba_ = mba;
	huffent* he = &hte_mba[m - 1];
	/* MBA */
	PUT_BITS(he->val, he->nb, nbb_, bb_, bs_);
	if (q != mquant_) {
		/* MTYPE = INTRA + TC + MQUANT */
		PUT_BITS(1, 7, nbb_, bb_, bs_);
		PUT_BITS(q, 5, nbb_, bb_, bs_);
		mquant_ = q;
	} else {
		/* MTYPE = INTRA + TC (no quantizer) */
		PUT_BITS(1, 4, nbb_, bb_, bs_);
	}

	/* luminance */
	encode_blk(blk + 0);
	encode_blk(blk + 64);
	encode_blk(blk + 128);
	encode_blk(blk + 192);
	/* chominance */
	encode_blk(blk + 256);
	encode_blk(blk + 320);
}

void H261Encoder::encode_frm(const u_char* lum, u_char* crvec)
{
	int step, boff;
	if (cif_) {
		step = 1;
		boff = 44;
	} else {
		step = 2;
		boff = 22;
	}

	for (int gob = 0; gob < ngob_; gob += step) {
		int loff = loff_[gob];
		int coff = coff_[gob];
		int blkno = blkno_[gob];

		/* GSC/GN */
		PUT_BITS(0x10 | (gob + 1), 20, nbb_, bb_, bs_);
		/* GQUANT/GEI */
		PUT_BITS(quant_ << 1, 6, nbb_, bb_, bs_);
		mquant_ = quant_;

		mba_ = 0;
		int line = 11;
		for (int mba = 1; mba <= 33; ++mba) {
			/*
			 * If the conditional replenishment algorithm
			 * has decided to send any of the blocks of
			 * this macroblock, code it.
			 */
			int s = crvec[blkno];
			s |= crvec[blkno + 1];
			s |= crvec[blkno + boff];
			s |= crvec[blkno + boff + 1];
			if (s & (CR_LQ|CR_HQ)) {
				crvec[blkno] = CR_BG;
				crvec[blkno + 1] = CR_BG;
				crvec[blkno + boff] = CR_BG;
				crvec[blkno + boff + 1] = CR_BG;
				encode_mb(mba, lum, loff, coff);
			}
			loff += 16;
			coff += 8;
			blkno += 2;
			if (--line <= 0) {
				line = 11;
				blkno += bstride_;
				loff += lstride_;
				coff += cstride_;
			}
		}
		/*
		 * Remember where the end of each GOB falls
		 * so we can do RTP fragmentation.
		 */
		offset_[gob].off = (bs_ - buffer_) + (nbb_ >> 3);
		offset_[gob].bit = nbb_ & 7;
	}
	/* flush bit buffer */
	STORE_BITS(bs_, bb_);
}

int H261Encoder::encode(n_long ts, const u_char* frame, u_char* crvec)
{
	flush();

	bs_ = buffer_;
	bb_ = 0;
	nbb_ = 0;

	/* PSC */
	PUT_BITS(0x0001, 16, nbb_, bb_, bs_);
	/* GOB 0 -> picture header */
	PUT_BITS(0, 4, nbb_, bb_, bs_);
	/* TR (XXX should do this right) */
	PUT_BITS(0, 5, nbb_, bb_, bs_);
	/* PTYPE = CIF */
	int pt = cif_ ? 4 : 0;
	PUT_BITS(pt, 6, nbb_, bb_, bs_);
	/* PEI */
	PUT_BITS(0, 1, nbb_, bb_, bs_);

	encode_frm(frame, crvec);
	
	/*
	 * We've encoded the entire frame above.  Now spin over
	 * the GOBs and fragment them as necessary.  We either
	 * fragment a single GOB into a number of packets, or we
	 * stuff multiple GOBs into a single packet.  We try to
	 * stuff extra GOBs in the last packet of a fragmented
	 * GOB, if they fit.
	 */
	int step = cif_ ? 1 : 2;
	int off = 0;
	int tot = 0;
	int sbit = 8;
	int mdu = mtu_ - HDRSIZE;
	for (int gob = 0; gob < ngob_; ) {
		/* strip GSC, which is exactly 16 bits */
		off += 2;

		int noff = offset_[gob].off;
		int nbit = offset_[gob].bit;
		gob += step;

		int cc = noff - off;
		if (nbit != 0)
			cc += 1;

		pktbuf* pb = alloch(ts);
		rtphdr* rh = (rtphdr*)pb->hdr;
		u_short flags = 0xc0 | cif_ | sbit << 12;

		/*
		 * Fragment this GOB if necessary.
		 */
		while (cc > mdu) {
			*(short*)(rh + 1) = htons(flags);
			pb->iov[0].iov_len = HDRSIZE;
			pb->iov[1].iov_base = (caddr_t)buffer_ + off;
			pb->iov[1].iov_len = mdu;
			Framer::send(pb);
			off += mdu;
			tot += mdu + HDRSIZE;
			cc -= mdu;

			pb = alloch(ts);
			rh = (rtphdr*)pb->hdr;
			flags = 0xc0 | cif_;	/*color, intra, size*/
		}
		/*
		 * Now see how many GOBs we can stuff in the rest of
		 * this packet.  The packet may be empty or it may be
		 * the last fragment of a fragmented GOB.
		 */
		while (gob < ngob_) {
			int toff = offset_[gob].off;
			int tbit = offset_[gob].bit;
			int tcc = toff - off;
			if (tbit != 0)
				tcc += 1;
			if (tcc > mdu)
				break;
			cc = tcc;
			noff = toff;
			nbit = tbit;
			gob += step;
		}
 		int ebit = (8 - nbit) & 7;
		flags |= (8 | ebit) << 8;
		*(short*)(rh + 1) = htons(flags);
		if (gob >= ngob_)
			rh->rh_flags |= htons(RTP_M);

		pb->iov[0].iov_len = HDRSIZE;
		pb->iov[1].iov_base = (caddr_t)buffer_ + off;
		pb->iov[1].iov_len = cc;
		Framer::send(pb);
		tot += cc + HDRSIZE;

		off = noff;
		sbit = 8 | nbit;
	}
	return (tot);
}
