/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef vic_framer_h
#define vic_framer_h

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include "net.h"
#include "packet.h"
#include "timer.h"

class Network;
class Encoder;
class Tcl;

/*
 * An `application-level framer'.  The base object for performing
 * the outbound path of the application level protocol.
 */
class Framer : public Timer {
protected:
	Framer(Network* net, int mtu);
public:
	virtual int command(Tcl&, int argc, char** argv);
	virtual void timeout();

	struct buffer {
		struct buffer* next;
		u_char data[VIC_MTU];
	};
	struct pktbuf {
		struct pktbuf* next;
		iovec iov[2];
		u_char hdr[MAXHDR];
		buffer* buf;
	};
	virtual ~Framer();
	static void dump(int fd);
	inline int format() const { return (format_); }
	inline void bps(int kbps) {
		pkttime_ = 1e6 * (8 * mtu_ / (1000. * kbps));
	}
	void flush();
protected:
	void update(int nbytes);
	void dump(int fd, iovec*, int iovel) const;
	void loopback(pktbuf*);
	void send(pktbuf*);
	void transmit(pktbuf* pb);

	/*
	 * Buffer allocation hooks.
	 */
	pktbuf* alloch(n_long);
	void release(pktbuf*);
	pktbuf* alloc(n_long);

	int format_;		/* video format type (rtp encoding) */
	int mtu_;		/* mtu of wire (as seen by application) */
	u_short seqno_;
	Network* net_;
	msghdr mh_;

	static int dumpfd_;	/* fd to dump packet stream to */

	int nf_;		/* number of frames sent */
	int nb_;		/* number of bytes sent */
	int np_;		/* number of packets sent */
	
	/* packet transmission queue */
	int busy_;
	double pkttime_;	/* usec between packets */
	double nextpkttime_;
	pktbuf* head_;
	pktbuf* tail_;

private:
	static pktbuf* freehdrs_;
	static buffer* freebufs_;
	static int nbufs_;
	static int nhdrs_;
};

inline Framer::pktbuf* Framer::alloch(n_long ts)
{
	pktbuf* pb = freehdrs_;
	if (pb != 0)
		freehdrs_ = pb->next;
	else {
		/*XXX grow exponentially*/
		pb = new pktbuf;
		++nhdrs_;
	}
	rtphdr* rh = (rtphdr*)pb->hdr;
	int flags = RTP_VERSION << 14 | format_;
	rh->rh_flags = ntohs(flags);
	rh->rh_seqno = htons(seqno_);
	++seqno_;
	rh->rh_ts = htonl(ts);
	rh->rh_ssrc = net_->srcid();

	pb->iov[0].iov_base = (char*)rh;
	pb->buf = 0;

	return (pb);
}

inline Framer::pktbuf* Framer::alloc(n_long ts)
{
	pktbuf* pb = alloch(ts);
	buffer* p = freebufs_;
	if (p != 0)
		freebufs_ = p->next;
	else
		p = new buffer;

	pb->buf = p;
	pb->iov[1].iov_base = (char*)p->data;

	return (pb);
}

/*
 * YuvEncoder's take 
 * Encode, frame (e.g., packetize), and transmit a video frame.
 * Image convention is YUV format, 4:2:2.  Width and height
 * are specified to constructor.
 * XXX mention ranges
 */
class YuvEncoder : public Framer {
 protected:
	YuvEncoder(Network*, int mtu);
 public:
	virtual ~YuvEncoder();
	virtual int encode(n_long ts, const u_char* frame, u_char* crvec) = 0;
	virtual void size(int width, int height) = 0;
protected:
	int rover_;
	int width_;
	int height_;
	int framesize_;		/* width x height */
};

class NullEncoder : public Framer {
public:
	NullEncoder(Network* net, int mtu);
	void send(iovec*);
};


#endif
