/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef vic_grabber_h
#define vic_grabber_h

extern "C" {
#include <tk.h>
}

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include "inet.h"
#include "packet.h"
#include "iohandler.h"
#include "timer.h"

#ifdef _AIX
extern "C" int strcasecmp (const char *, const char *);
#endif

class Grabber;
class Framer;
class Network;
class JpegFramer;
class H261Framer;
class YuvEncoder;

/*
 * Dummy formats
 */
#define RTPCONT_NONE		-1

/*
 * Device template for the video capture hardware.
 * Device::instance returns the detected device.  When multiple
 * devices are present, the are distinguished by the device
 * specific type name.  A grabber is actually allocated by one
 * of the *_grabber methods.  The default allocation
 * methods print an incompatibility error and exit.
 */
class Device {
 public:
	Device(const char* name);
	static void init();
	static Device& lookup(const char* type);
	static Device& offline(const char* file);

	virtual int match(const char* type) const = 0;
	int compat(int fmt) const;
	int supports(const char* attr) const;
	int default_format() const { return (default_format_); }
	virtual Grabber* jpeg_grabber(JpegFramer*);
	/* 4:2:2 YUV, any geometry */
	virtual Grabber* yuv_grabber(YuvEncoder*);
	/* 4:1:1 YUV, CIF (352x288) or QCIF (176x144) */
	virtual Grabber* cif_grabber(YuvEncoder*);
	virtual Grabber* h261_grabber(H261Framer*);
#ifdef notyet
	virtual Grabber* mpeg_grabber(MPEGFramer*);
	virtual Grabber* cellb_grabber(CellBFramer*);
#endif
	inline const char* name() const { return (name_); }
 private:
	static int device_cmd(ClientData, Tcl_Interp*, int argc, char **argv);
	static int encoder_cmd(ClientData, Tcl_Interp*, int argc, char **argv);
	static int open(const char* type, const char* fmt);
	static Device* alldev_;
	static Device* nulldev_;
	static Grabber* grabber_;
	static Framer* framer_;
	Device* next_;
	const char* name_;
 protected:
	int default_format_;
	u_long attrs_;
#define DEV_ATTR_RAW		0x01
#define DEV_ATTR_CIF		0x02
#define DEV_ATTR_HW_JPEG	0x04
#define DEV_ATTR_HW_H261	0x08
#define DEV_ATTR_HW_MPEG	0x10
#define DEV_ATTR_HW_CELLB	0x20
#define DEV_ATTR_SMALL		0x010000
#define DEV_ATTR_LARGE		0x020000
#define DEV_ATTR_PORT		0x040000
#define DEV_ATTR_TYPE		0x080000
};

class Grabber : public Timer {
 protected:
	Grabber();
 public:
	virtual ~Grabber();
	virtual int command(int argc, char** argv);
	virtual void fps(int);
	inline int fps() const { return (fps_); }
	virtual void bps(int);
	inline int bps() const { return (bps_); }
	inline void locked(int v) { locked_ = v; }
	inline void fillrate(int v) { idle_high_ = v; }
	double tick(int n);
	virtual void start();
	virtual void stop();
	inline int status() const { return (status_); }

	virtual int grab();
	virtual void timeout();
	int background() const;
 protected:
	n_long ntptime();

	/* hooks for conditional replenishment algorithm */
	void suppress(const u_char* devbuf, const u_char* lumbuf);
	void saveblks(u_char* lum);
	void crinit(int w, int h, int needcache = 1);
	void idlefill();
	double advance();

	u_char* crvec_;
	u_char* cache_;/*XXX*/
	int blkw_;
	int blkh_;
	int scan_;
	int nblk_;
	int stride_;
	int rover_;
	int idle_low_;
	int idle_high_;

	double frametime_;	/* 1 / fps */
	double frameclock_;	/* logical clock for pumping out frames */

	int fps_;		/* desired frames per second */
	int bps_;		/* desired bits per second */
	int running_;		/* true when grabber active */
	int locked_;		/* disables remote control */
	int status_;		/* true if grabber okay */

	double delta_;		/* most recent inter-frame time */
};

#define SUPPRESS_BODY(devbuf, lumbuf, scale) \
	int off = scan_ * stride_; \
	const u_char* in = (devbuf) + (off * (scale)); \
	const u_char* lum = (lumbuf) + off; \
 \
	int w = blkw_; \
	u_char* crv = crvec_; \
	int stride = 7 * stride_; \
	for (int y = 0; y < blkh_; y++) { \
		for (int x = 0; x < w; x++) { \
			int pri = diff(in, lum); \
			if (pri >= 64) { \
				crv[0] = CR_LQ; \
				if (y > 0) { \
					crv[-w] = CR_LQ; \
					if (x > 0) \
						crv[-w-1] = CR_LQ; \
					if (x < w - 1) \
						crv[-w+1] = CR_LQ; \
				} \
				if (y < blkh_ - 1) { \
					crv[w] = CR_LQ; \
					if (x > 0) \
						crv[w-1] = CR_LQ; \
					if (x < w - 1) \
						crv[w+1] = CR_LQ; \
				} \
				if (x > 0) \
					crv[-1] = CR_LQ; \
				if (x < w - 1) \
					crv[1] = CR_LQ; \
			} else { \
				/* \
				 * There was no motion.  Age this block. \
				 * Once we hit the age threshold, we \
				 * will set CR_HQ as a hint to send a \
				 * high-quality version of the block. \
				 * After this the block will stop aging, \
				 * until there is motion (it still can \
				 * be sent as background fill). \
				 */ \
				int s = *crv; \
				if ((s & (CR_HQ|CR_LQ|CR_BG)) == 0) { \
					s += 1; \
					if (s >= CR_AGETHRESH) \
						s |= CR_HQ; \
					*crv = s; \
				} \
			} \
			lum += 8; \
			in += 8 * scale; \
			++crv; \
		} \
		lum += stride; \
		in += stride * scale; \
	} \
	/* \
	 * Run background fill algorithm. \
	 */ \
	idlefill();

#endif
