/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/*
 * This code is derived from the Independent JPEG Group's JPEG software:
 *
 * Copyright (C) 1991, 1992, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying
 * README.IJPG file.
 */

#include "jpeg.h"
#include "endian.h"
#include "dct.h"

#include <stdlib.h>
#include <stdio.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <string.h>

class JpegDecoder_411 : public JpegDecoder {
 public:
	JpegDecoder_411(const config&);
	virtual int decode(u_char* in, int len);
};

class JpegDecoder_422 : public JpegDecoder {
 public:
	JpegDecoder_422(const config&);
	virtual int decode(u_char* in, int len);
};

JpegDecoder_411::JpegDecoder_411(const config& c) : JpegDecoder(c, 411)
{
}

JpegDecoder_422::JpegDecoder_422(const config& c) : JpegDecoder(c, 422)
{
}

JpegDecoder* JpegDecoder::create(const config& c)
{
	if (c.ncomp == 3 && c.comp[0].hsf == 2 &&
	    c.comp[1].hsf == 1 && c.comp[1].vsf == 1 &&
	    c.comp[2].hsf == 1 && c.comp[2].vsf == 1) {	
		if (c.comp[0].vsf == 2)
			return (new JpegDecoder_411(c));
		if (c.comp[0].vsf == 1)
			return (new JpegDecoder_422(c));
	}
	return (0);
}

int quality_to_qfactor(int v)
{
	if (v < 1)
		v = 5000;
	else if (v < 50)
		v = 5000 / v;
	else if (v < 100)
		v = 200 - v * 2;
	else 
		v = 1;

	return (v);
}

/*
 * Table K.1 from JPEG spec.
 */
static const int jpeg_luma_quantizer[64] = {
	16, 11, 10, 16, 24, 40, 51, 61,
	12, 12, 14, 19, 26, 58, 60, 55,
	14, 13, 16, 24, 40, 57, 69, 56,
	14, 17, 22, 29, 51, 87, 80, 62,
	18, 22, 37, 56, 68, 109, 103, 77,
	24, 35, 55, 64, 81, 104, 113, 92,
	49, 64, 78, 87, 103, 121, 120, 101,
	72, 92, 95, 98, 112, 100, 103, 99
};

/*
 * Table K.2 from JPEG spec.
 */
static const int jpeg_chroma_quantizer[64] = {
	17, 18, 24, 47, 99, 99, 99, 99,
	18, 21, 26, 66, 99, 99, 99, 99,
	24, 26, 56, 99, 99, 99, 99, 99,
	47, 66, 99, 99, 99, 99, 99, 99,
	99, 99, 99, 99, 99, 99, 99, 99,
	99, 99, 99, 99, 99, 99, 99, 99,
	99, 99, 99, 99, 99, 99, 99, 99,
	99, 99, 99, 99, 99, 99, 99, 99,
};

void jpeg_qt(int q, int* out, const int* in)
{
	q = quality_to_qfactor(q);
	for (int i = 0; i < 64; i++) {
		int val = (q * in[i] + 50) / 100;
		/*
		 * Baseline JPEG restricts range to [1,255].
		 */
		if (val < 1)
			val = 1;
		else if (val > 255)
			val = 255;
		out[i] = val;
	}
}

void jpeg_luma_qt(int q, int* qt)
{
	jpeg_qt(q, qt, jpeg_luma_quantizer);
}

void jpeg_chroma_qt(int q, int* qt)
{
	jpeg_qt(q, qt, jpeg_chroma_quantizer);
}

static const unsigned char dc_luminance_bits[17] =
    { /* 0-base */ 0, 0, 1, 5, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0 };
static const unsigned char dc_luminance_val[] =
    { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };
  
static const unsigned char dc_chrominance_bits[17] =
    { /* 0-base */ 0, 0, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0 };
static const unsigned char dc_chrominance_val[] =
    { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };
  
static const unsigned char ac_luminance_bits[17] =
    { /* 0-base */ 0, 0, 2, 1, 3, 3, 2, 4, 3, 5, 5, 4, 4, 0, 0, 1, 0x7d };
static const unsigned char ac_luminance_val[] =
    { 0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12,
      0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07,
      0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xa1, 0x08,
      0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52, 0xd1, 0xf0,
      0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
      0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28,
      0x29, 0x2a, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
      0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
      0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
      0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
      0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
      0x7a, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
      0x8a, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
      0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
      0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
      0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5,
      0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4,
      0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xe1, 0xe2,
      0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
      0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
      0xf9, 0xfa };
  
static const unsigned char ac_chrominance_bits[17] =
    { /* 0-base */ 0, 0, 2, 1, 2, 4, 4, 3, 4, 7, 5, 4, 4, 0, 1, 2, 0x77 };
static const unsigned char ac_chrominance_val[] =
    { 0x00, 0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21,
      0x31, 0x06, 0x12, 0x41, 0x51, 0x07, 0x61, 0x71,
      0x13, 0x22, 0x32, 0x81, 0x08, 0x14, 0x42, 0x91,
      0xa1, 0xb1, 0xc1, 0x09, 0x23, 0x33, 0x52, 0xf0,
      0x15, 0x62, 0x72, 0xd1, 0x0a, 0x16, 0x24, 0x34,
      0xe1, 0x25, 0xf1, 0x17, 0x18, 0x19, 0x1a, 0x26,
      0x27, 0x28, 0x29, 0x2a, 0x35, 0x36, 0x37, 0x38,
      0x39, 0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48,
      0x49, 0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
      0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
      0x69, 0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78,
      0x79, 0x7a, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
      0x88, 0x89, 0x8a, 0x92, 0x93, 0x94, 0x95, 0x96,
      0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5,
      0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4,
      0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3,
      0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2,
      0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda,
      0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
      0xea, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
      0xf9, 0xfa };

void JpegDecoder::defaults(JpegDecoder::config& c)
{
	c.width = 0;
	c.height = 0;

	c.precision = 8;
	c.ncomp = 3;
	/* Y */
	c.comp[0].id = 0;
	c.comp[0].hsf = 2;
	c.comp[0].vsf = 1;
	c.comp[0].qno = 0;
	/* U */
	c.comp[1].id = 1;
	c.comp[1].hsf = 1;
	c.comp[1].vsf = 1;
	c.comp[1].qno = 1;
	/* V */
	c.comp[2].id = 2;
	c.comp[2].hsf = 1;
	c.comp[2].vsf = 1;
	c.comp[2].qno = 1;
	
	c.dc_huffbits[0] = dc_luminance_bits;
	c.dc_huffval[0] = dc_luminance_val;
	c.dc_huffbits[1] = dc_chrominance_bits;
	c.dc_huffval[1] = dc_chrominance_val;
	c.ac_huffbits[0] = ac_luminance_bits;
	c.ac_huffval[0] = ac_luminance_val;
	c.ac_huffbits[1] = ac_chrominance_bits;
	c.ac_huffval[1] = ac_chrominance_val;
	for (int i = 2; i < 4; ++i) {
		c.ac_huffbits[i] = 0;
		c.ac_huffval[i] = 0;
		c.dc_huffbits[i] = 0;
		c.dc_huffval[i] = 0;
	}
	c.comp[0].dc_tbl_no = 0;
	c.comp[0].ac_tbl_no = 0;
	c.comp[1].dc_tbl_no = 1;
	c.comp[1].ac_tbl_no = 1;
	c.comp[2].dc_tbl_no = 1;
	c.comp[2].ac_tbl_no = 1;
}

/*
 * Set the quantizer for this configuration.
 * Q is the IJPG quality factor, which has a value in [0,100].
 */
void JpegDecoder::quantizer(JpegDecoder::config& c, int q)
{
	jpeg_luma_qt(q, c.qtab[0]);
	jpeg_chroma_qt(q, c.qtab[1]);
}

int JpegDecoder::q_to_thresh(int q)
{
	int s = q / 10;
	if (q <= 30 || q >= 80)
		s += 1;
	return (s);
}

JpegDecoder::JpegDecoder(const config& c, int decimation)
	: decimation_(decimation), color_(1), width_(-1), height_(-1),
	cache_(0), thresh_(0), cthresh_(6), renderer_(0)
{
	for (int i = NUM_HUFF_TBLS; --i >= 0; ) {
		dcht_[i] = 0;
		acht_[i] = 0;
	}
	memset((char*)comp_, 0, sizeof(comp_));
	init(c);
}

JpegDecoder::~JpegDecoder()
{
	freehufftab();
	delete comp_[0].frm;
	delete cache_;
}

void JpegDecoder::init(const config& c)
{
	rlen_ = 0;

	memcpy((char*)qt_, (char*)c.qtab, sizeof(qt_));
	int i;
	for (i = 0; i < 4; ++i)
		rdct_fold_q(qt_[i], fqt_[i]);

	ncomp_ = c.ncomp;/*XXX*/
	if (width_ != c.width || height_ != c.height) {
		width_ = c.width;
		height_ = c.height;

		/*
		 * Allocate the frame store
		 */
		delete comp_[0].frm;
		int n, off1, off2;
		size_ = width_ * height_;
		off1 = size_;
		if (decimation_ == 422) {
			n = size_ * 2;
			off2 = off1 + size_ / 2;
		} else if (decimation_ == 411) {
			/* 411 */
			n = size_ + size_ / 2;
			off2 = off1 + size_ / 4;
		} else {
			fprintf(stderr,
				"JpegDecoder: unsuppored decimation\n");
			exit(1);
		}
		u_char* p = new u_char[n];
		comp_[0].frm = p;
		comp_[1].frm = p + off1;
		comp_[2].frm = p + off2;

		/* Initialize to gray */
		memset((char*)p, 0x80, n);
	}
	for (i = ncomp_; --i >= 0; ) {
		int id = c.comp[i].id;
		comp_[id].hsf = c.comp[i].hsf;
		comp_[id].vsf = c.comp[i].vsf;
		comp_[id].qno = c.comp[i].qno;
		comp_[id].dc_tbl_no = c.comp[i].dc_tbl_no;
		comp_[id].ac_tbl_no = c.comp[i].ac_tbl_no;
	}
	/*
	 * XXX should check if huffman table won't change
	 * before reallocating.
	 */
	freehufftab();
	for (i = 0; i < 4; ++i) {
		if (c.dc_huffval[i] != 0) {
			int id = c.comp[i].id;
			dcht_[id] = huffbuild(c.dc_huffbits[i],
					      c.dc_huffval[i]);
		}
		if (c.ac_huffval[i] != 0) {
			int id = c.comp[i].id;
			acht_[id] = huffbuild(c.ac_huffbits[i],
					      c.ac_huffval[i]);
		}
	}

	int maxh = 1;
	int maxv = 1;
	for (i = ncomp_; --i >= 0; ) {
		if (maxh < comp_[i].hsf)
			maxh = comp_[i].hsf;
		if (maxv < comp_[i].vsf)
			maxv = comp_[i].vsf;
	}
	ncol_ = (width_ + 8 * maxh - 1) / (8 * maxh);
	nrow_ = (height_ + 8 * maxv - 1) / (8 * maxv);

	int nmcu = ncol_ * nrow_;
	int nblk = nmcu * comp_[0].hsf * comp_[0].vsf;
	nblk += nmcu * comp_[1].hsf * comp_[1].vsf;
	nblk += nmcu * comp_[2].hsf * comp_[2].vsf;
#define NCC 6
	int ns = NCC * nblk;
	delete cache_;
	cache_ = new short[ns];
	/*
	 * Initialize to some "large value" so threshold
	 * is exceeded on first pass.
	 */
	memset(cache_, 0x7f, ns * sizeof(*cache_));
}

void JpegDecoder::freehufftab()
{
	for (int i = 0; i < 4; ++i) {
		if (dcht_[i] != 0)
			free(dcht_[i]);
		if (acht_[i] != 0)
			free(acht_[i]);
	}
}

#if NCC != 6

@BUG in blkdiff@
#endif
#define ABS(t) ((t) - (((t) >> 31 & (t)) << 1))
inline int blkdiff(short* blk, short* cache)
{
	int t = blk[0] - cache[0];
	int d = ABS(t);
	t = blk[1] - cache[1];
	d += ABS(t);
	t = blk[2] - cache[2];
	d += ABS(t);
	t = blk[3] - cache[3];
	d += ABS(t);
	t = blk[4] - cache[4];
	d += ABS(t);
	t = blk[5] - cache[5];
	d += ABS(t);
	return (d);
}

#ifdef notdef
/* XXX this is negligibly faster than above, and has a bug */
inline int blkdiff(short* blk, short* cache)
{
	u_int* p0 = (u_int*)blk;
	u_int* p1 = (u_int*)cache;
	
	int m = ~0x80008000;
	int t = (p0[0] >> 1) & m;
	t += (~p1[0] >> 1) & m;
	int v = t << 17 >> 17;
	int d = ABS(v);
	t = (t << 1) >> 17;
	d += ABS(v);

	t = (p0[1] >> 1) & m;
	t += (~p1[1] >> 1) & m;
	v = t << 17 >> 17;
	d += ABS(v);
	t = (t << 1) >> 17;
	d += ABS(v);

	t = (p0[2] >> 1) & m;
	t += (~p1[2] >> 1) & m;
	v = t << 17 >> 17;
	d += ABS(v);
	t = (t << 1) >> 17;
	d += ABS(v);

	return (d);
}
#endif

int JpegDecoder::rdqt(const u_char* p)
{
	int len = *p++ << 8;
	len |= *p++;
  
	const u_char* ep = p + len - 2;
	while (p < ep) {
		int n = *p++;
		int prec = n >> 4;
		n &= 0x0f;
		if (n >= 4) {
			/*XXX illegal number*/
			return (-1);
		}
		int i;
		int qt[64];
		for (i = 0; i < 64; i++) {
			int v = *p++;
			if (prec)
				v = (v << 8) + *p++;
			qt[i] = v;
		}
		for (i = 0; i < 64; i++) {
			/* Compute new table only if it changed */
			int j = ROWZAG[i];
			if (qt[i] != qt_[n][j]) {
				for (; i < 64; ++i) {
					j = ROWZAG[i];
					qt_[n][j] = qt[i];
				}
				rdct_fold_q(qt_[n], fqt_[n]);
				break;
			}
		}
	}
	return (len);
}

void JpegDecoder::restart()
{
	int c;
	nbb_ = 0;
	/*XXXwhat if ff is sitting in bit buffer?*/
	/* Scan for next JPEG marker */
	do {
		do {			/* skip any non-FF bytes */
			c = *inb_++;
		} while (c != 0xFF);
		do {
			/* skip any duplicate FFs */
			/* we don't increment nbytes here since extra FFs are legal */
			c = *inb_++;
		} while (c == 0xFF);
	} while (c == 0);		/* repeat if it was a stuffed FF/00 */
#ifdef notdef
	if (nbytes != 1)
		WARNMS2(cinfo->emethods,
			"Corrupt JPEG data: %d extraneous bytes before marker 0x%02x",
			nbytes-1, c);
	
#endif
#ifdef notdef
	if (c != (RST0 + cinfo->next_restart_num)) {
		/* Uh-oh, the restart markers have been messed up too. */
		/* Let the file-format module try to figure out how to resync. */
		(*cinfo->methods->resync_to_restart) (cinfo, c);
	} else
		TRACEMS1(cinfo->emethods, 2, "RST%d", cinfo->next_restart_num);
#endif
	/* Re-initialize DC predictions to 0 */
	comp_[0].dc = 0;
	comp_[1].dc = 0;
	comp_[2].dc = 0;
#ifdef notdef
	cinfo->next_restart_num = (cinfo->next_restart_num + 1) & 7;
#endif
}

u_char* JpegDecoder::parseJFIF(u_char* in)
{
	int t;
	while (in < end_) {
		if (*in++ != 0xff)
			continue;
		/*XXX need more checks for buffer overflow*/
		switch (*in++) {

		default:
			/* Don't know.  Keep looking for SOS. */
			continue;
			
		case 0xdb:
			/* quantization table */
			t = rdqt(in);
			if (t < 0)
				/*XXX*/
				return (end_);
			in += t;
			continue;

		case 0xdd:
			/* restart interval definition */
			t = *in++ << 8;
			t |= *in++;
			if (t != 4)
				/* XXX bad length */
				;
			rlen_ = *in++ << 8;
			rlen_ |= *in++;
			rcnt_ = 0;
			continue;

		case 0xda:
			/* start-of-scan marker */
			if (in + 2 <= end_) {
				/* skip over SOS */
				int t = *in++ << 8;
				t |= *in++;
				in += (t - 2);
			}
			return (in);
		}
	}
	/*XXX*/
	return (end_);
}

/*XXX*/
#ifdef INT_64
#define MASK_DECL	INT_64 m0
#define MASK_VAL	m0
#define MASK_REF	&m0
inline void JpegDecoder::rdct(int nc, register short *bp, INT_64 mask,
			      u_char* pixels, int stride, int qno)
{
#ifdef notyet
	if (nc == 2) {
		for (int k = 1; k < 64; ++k) 
			if (mask & ((INT_64)1 << k)) {
				int s = bp[k] * qt_[qno][k];
				int dc = (bp[0] * qt_[qno][0] >> 3) + 128;
				/*XXX limit dc?*/
				bv_rdct1(dc, bp, k, pixels, stride);
				return;
			}
	}
#endif
	if (nc != 0)
		::rdct(bp, mask, pixels, stride, fqt_[qno]);
#ifdef notdef
	if (nc > cthresh_)
		::rdct(bp, mask, pixels, stride, fqt_[qno]);
	else if (nc > 0)
		bv_rdct(bp, mask, pixels, stride, qt_[qno]);
#endif
}
#else
#define MASK_DECL	u_int mask[2]
#define MASK_VAL	mask[0], mask[1]
#define MASK_REF	mask
inline void JpegDecoder::rdct(int nc, register short *bp, u_int m0, u_int m1,
			      u_char* pixels, int stride, int qno)
{
#ifdef notyet
	if (nc > cthresh_)
		::rdct(bp, m0, m1, pixels, stride, fqt_[qno]);
	else if (nc > 0)
		bv_rdct(bp, m0, m1, pixels, stride, qt_[qno]);
#else
	if (nc != 0)
		::rdct(bp, m0, m1, pixels, stride, fqt_[qno]);
#endif
}
#endif

int JpegDecoder_422::decode(u_char* in, int len)
{
	minx_ = ncol_;
	miny_ = nrow_;
	maxx_ = 0;
	maxy_ = 0;

	inb_ = in;
	end_ = in + len;
	nbb_ = 0;
	/*
	 * If first symbol is a marker (a not a stuffed ff),
	 * assume a jfif header is present and parse it.
	 * XXX this could change state that needs to be
	 * communicated back to caller.
	 */
	if (in[0] == 0xff && in[1] != 0)
		inb_ = parseJFIF(inb_);

	u_char* p0 = comp_[0].frm;
	u_char* p1 = comp_[1].frm;

	huffreset();
	short* cache = cache_;
	u_long ablk[sizeof(short) * 64 / sizeof(u_long)];
	short* blk = (short*)ablk;

	JpegRenderer* renderer = renderer_;
	int q0 = comp_[0].qno;
	int q1 = comp_[1].qno;
	for (int y = 0; y < nrow_; ++y) {
		for (int x = 0; x < ncol_; ++x) {
			MASK_DECL;
			/*
			 * If we're handling restart markers,
			 * check if we need to resync.
			 */
			if (rlen_ != 0 && --rcnt_ <= 0) {
				rcnt_ = rlen_;
				restart();
			}

			int nc = huffparse(comp_[0], blk, cache, MASK_REF);
			int dontskip = nc;
			cache += NCC;
			rdct(nc, blk, MASK_VAL, p0, width_, q0);
			nc = huffparse(comp_[0], blk, cache, MASK_REF);
			dontskip |= nc;
			cache += NCC;
			rdct(nc, blk, MASK_VAL, p0 + 8, width_, q0);
			p0 += 16;
			if (color_) {
				/*
				 * If we found above that the luminance
				 * planes exceeded the threhold, decode
				 * the choma planes unconditionally.
				 * Otherwise, see if they can be
				 * suppressed too.
				 */
				nc = huffparse(comp_[1], blk, cache,
					       MASK_REF, dontskip);
				dontskip |= nc;
				cache += NCC;
				rdct(nc, blk, MASK_VAL, p1, width_ / 2, q1);
				nc = huffparse(comp_[2], blk, cache,
					       MASK_REF, dontskip);
				dontskip |= nc;
				cache += NCC;
				rdct(nc, blk, MASK_VAL, p1 + size_ / 2,
				     width_ / 2, q1);

				p1 += 8;
			} else {
				(void)huffskip(comp_[1]);
				(void)huffskip(comp_[2]);
				cache += 2 * NCC;
			}
			/* Update bounding box */
			/* XXX these can be locals */
			if (dontskip) {
				if (x < minx_)
					minx_ = x;
				if (x > maxx_)
					maxx_ = x;
				if (y < miny_)
					miny_ = y;
				if (y > maxy_)
					maxy_ = y;
				if (renderer)
					renderer->render(x << 4, y << 3,
							 16, 8,
							 comp_[0].frm,
							 comp_[1].frm,
							 comp_[2].frm);
			}
		}
		/*
		 * We're at the end of the current line.
		 * Back up to the beggining, then skip down
		 * one row to the start of the next mcu.
		 */
		p0 -= width_;
		p0 += 8 * width_;
		p1 -= width_ / 2;
		p1 += 8 * width_ / 2;
	}
	/*
	 * Convert mcu coords to pixel coords.
	 * A 4:2:2 mcu is 16x8.
	 */
	minx_ <<= 4;
	maxx_ <<= 4;
	miny_ <<= 3;
	maxy_ <<= 3;
	maxx_ += 16;
	maxy_ += 8;

	return (0);
}

int JpegDecoder_411::decode(u_char* in, int len)
{
	minx_ = ncol_;
	miny_ = nrow_;
	maxx_ = 0;
	maxy_ = 0;

	inb_ = in;
	end_ = in + len;
	nbb_ = 0;
	/*
	 * If first symbol is a marker (a not a stuffed ff),
	 * assume a jfif header is present and parse it.
	 * XXX this could change state that needs to be
	 * communicated back to caller.
	 */
	if (in[0] == 0xff && in[1] != 0)
		inb_ = parseJFIF(inb_);

	u_char* p0 = comp_[0].frm;
	u_char* p1 = comp_[1].frm;
	huffreset();
	short* cache = cache_;
	u_long ablk[sizeof(short) * 64 / sizeof(u_long)];
	short* blk = (short*)ablk;

	JpegRenderer* renderer = renderer_;
	int q0 = comp_[0].qno;
	int q1 = comp_[1].qno;
	for (int y = 0; y < nrow_; ++y) {
		for (int x = 0; x < ncol_; ++x) {
			MASK_DECL;
			/*
			 * If we're handling restart markers,
			 * check if we need to resync.
			 */
			if (rlen_ != 0 && --rcnt_ <= 0) {
				rcnt_ = rlen_;
				restart();
			}
			int nc = huffparse(comp_[0], blk, cache, MASK_REF);
			int dontskip = nc;
			cache += NCC;
			rdct(nc, blk, MASK_VAL, p0, width_, q0);
			nc = huffparse(comp_[0], blk, cache, MASK_REF);
			dontskip |= nc;
			cache += NCC;
			rdct(nc, blk, MASK_VAL, p0 + 8, width_, q0);
			nc = huffparse(comp_[0], blk, cache, MASK_REF);
			dontskip |= nc;
			cache += NCC;
			rdct(nc, blk, MASK_VAL, p0 + 8 * width_, width_, q0);
			nc = huffparse(comp_[0], blk, cache, MASK_REF);
			dontskip |= nc;
			cache += NCC;
			rdct(nc, blk, MASK_VAL, p0 + 8 * width_ + 8, width_, q0);
			p0 += 16;
			if (color_) {
				/*
				 * If we found above that the luminance
				 * planes exceeded the threhold, decode
				 * the choma planes unconditionally.
				 * Otherwise, see if they can be
				 * suppressed too.
				 */
				nc = huffparse(comp_[1], blk, cache,
					       MASK_REF, dontskip);
				dontskip |= nc;
				cache += NCC;
				rdct(nc, blk, MASK_VAL, p1, width_ / 2, q1);
				nc = huffparse(comp_[2], blk, cache,
					       MASK_REF, dontskip);
				cache += NCC;
				dontskip |= nc;
				rdct(nc, blk, MASK_VAL,
				     p1 + size_ / 4, width_ / 2, q1);
				p1 += 8;
			} else {
				(void)huffskip(comp_[1]);
				(void)huffskip(comp_[2]);
			}
			/* Update bounding box */
			/* XXX these can be locals */
			if (dontskip) {
				if (x < minx_)
					minx_ = x;
				if (x > maxx_)
					maxx_ = x;
				if (y < miny_)
					miny_ = y;
				if (y > maxy_)
					maxy_ = y;
				if (renderer)
					renderer->render(x << 4, y << 4,
							 16, 16,
							 comp_[0].frm,
							 comp_[1].frm,
							 comp_[2].frm);
			}
		}
		p0 -= width_;
		p0 += 16 * width_;
		p1 -= width_ / 2;
		p1 += 8 * width_ / 2;
	}
	/*
	 * Convert mcu coords to pixel coords.
	 * A 4:1:1 mcu is 16x16.
	 */
	minx_ <<= 4;
	maxx_ <<= 4;
	miny_ <<= 4;
	maxy_ <<= 4;
	maxx_ += 16;
	maxy_ += 16;

	return (0);
}

/* Figure F.12: extend sign bit */

#ifdef notdef
#define huff_EXTEND(x,s)  ((x) < extend_test[s] ? (x) + extend_offset[s] : (x))

static const int extend_test[16] =   /* entry n is 2**(n-1) */
  { 0, 0x0001, 0x0002, 0x0004, 0x0008, 0x0010, 0x0020, 0x0040, 0x0080,
    0x0100, 0x0200, 0x0400, 0x0800, 0x1000, 0x2000, 0x4000 };

static const int extend_offset[16] = /* entry n is (-1 << n) + 1 */
  { 0, ((-1)<<1) + 1, ((-1)<<2) + 1, ((-1)<<3) + 1, ((-1)<<4) + 1,
    ((-1)<<5) + 1, ((-1)<<6) + 1, ((-1)<<7) + 1, ((-1)<<8) + 1,
    ((-1)<<9) + 1, ((-1)<<10) + 1, ((-1)<<11) + 1, ((-1)<<12) + 1,
    ((-1)<<13) + 1, ((-1)<<14) + 1, ((-1)<<15) + 1 };
#else
/* is this really faster? */
inline int
huff_EXTEND(int x, int s)
{
	register int b = x >> (s - 1);
	register int m = ((b & 1) - 1) << s;
	return ((x | m) + (~b & 1));
}
#endif

/*
 * Read the next 16 bits off the bit string into the bit buffer.
 * Skip over zero-stuffed ff's but make no attempt to verify
 * that they aren't some other marker (which shouldn't be in the
 * middle of a block anyway).
 */
#define HUFFRQ(bb) \
 { \
	register int v; \
	register u_char *cp = inb_; \
 \
	bb <<= 16; \
	v = *cp++; \
	if (v == 0xff) ++cp; \
	bb |= v << 8; \
	v = *cp++; \
	if (v == 0xff) ++cp; \
	bb |= v; \
	inb_ = cp; \
 \
}

#define MASK(s) ((1 << (s)) - 1)

#define HUFF_DECODE(ht, nbb, bb, result) { \
	register int s_, v_; \
 \
	if (nbb < 16) { \
		HUFFRQ(bb); \
		nbb += 16; \
	} \
	v_ = (bb >> (nbb - 16)) & 0xffff; \
	s_ = (ht)[v_]; \
	nbb -= (s_ >> 8); \
	result = s_ & 0xff; \
 }

#define GET_BITS(n, nbb, bb, result) \
{ \
	nbb -= n; \
	if (nbb < 0)  { \
		HUFFRQ(bb); \
		nbb += 16; \
	} \
	(result) = ((bb >> nbb) & MASK(n)); \
}

#define SKIP_BITS(n, nbb, bb) \
{ \
	nbb -= n; \
	if (nbb < 0)  { \
		HUFFRQ(bb); \
		nbb += 16; \
	} \
}

int JpegDecoder::huffdc(component& p)
{
	/* Decode a single block's worth of coefficients */
		
	/* Section F.2.2.1: decode the DC coefficient difference */
	register int bb = bb_;
	register int nbb = nbb_;
	u_short* ht = dcht_[p.dc_tbl_no];
	register int s, r;
	HUFF_DECODE(ht, nbb, bb, s);
	if (s != 0) {
		GET_BITS(s, nbb, bb, r);
		s = huff_EXTEND(r, s);
	}
	/* Convert DC difference to actual value, update predictor */
	s += p.dc;
	p.dc = s;
	
	/* Section F.2.2.2: decode the AC coefficients */
	ht = acht_[p.ac_tbl_no];
	for (register int k = 1; k < 64; ) {
		/* Symbol-1 */
		register int v;
		HUFF_DECODE(ht, nbb, bb, v);
		s = v & 15;
		r = v >> 4;
		if (s != 0) {
			k += r;
			/* Symbol-2 */
			SKIP_BITS(s, nbb, bb);
			++k;
		} else {
			if (r != 15)
				/* end of block */
				break;
			k += 16;
		}
	}
	nbb_ = nbb;
	bb_ = bb;

	return (0);
}

/*
 * Parse a huffman-encoded 8x8 block.  Blocks are independent
 * of eachother, except for the dc predictor, and blocks can
 * start and end on arbitrary bit boundaries.  No markers should
 * appear in the bit stream, and ff bytes should be zero stuffed
 * (i.e., replaced with ff 00).
 *
 * The block is coded as a sequence of pairs of symbols.  Where the
 * first symbol is a huffman-encoded value <r,n> where r is a four-bit
 * runlength and n is the length of the second symbol (of the pair),
 * which follows verbatim in the bit string.
 */
#ifdef INT_64
int JpegDecoder::huffparse(component& p, short* blk, short* cache,
			   INT_64* mask, int dontskip)
#else
int JpegDecoder::huffparse(component& p, short* blk, short* cache,
			   u_int* mask, int dontskip)
#endif
{
	register int bb = bb_;
	register int nbb = nbb_;
	u_short* ht = dcht_[p.dc_tbl_no];
	register int s, r;
	HUFF_DECODE(ht, nbb, bb, s);
	if (s != 0) {
		GET_BITS(s, nbb, bb, r);
		s = huff_EXTEND(r, s);
	}

	/* update predictor */
	s += p.dc;
	p.dc = s;
	blk[0] = s;
	int nc = 1;

#if NCC != 6
@BUG@
#endif
	u_int sblk[3];
#if BYTE_ORDER == LITTLE_ENDIAN
	sblk[0] = s;
#else
	sblk[0] = s << 16;
#endif
	sblk[1] = 0;
	sblk[2] = 0;

	/*
	 * First, grab only a few low frequency coefficients.
	 * If they aren't sufficiently different from the current
	 * block, skip over this block quickly.
	 */
	ht = acht_[p.ac_tbl_no];
	register int k = 1;
#ifdef INT_64
	INT_64 m0 = 1;
#else
	u_int m1 = 0;
	u_int m0 = 1;
#endif
	for (;;) {
		register int v;
		/* Symbol-1 */
		HUFF_DECODE(ht, nbb, bb, v);
		s = v & 15;
		r = v >> 4;
		if (s != 0) {
			k += r;
			if (k >= 6) {
				k -= r;
				break;
			}
			/* Symbol-2 */
			GET_BITS(s, nbb, bb, v);
			s = huff_EXTEND(v, s);
			((short*)sblk)[k] = s;/*XXX*/
			v = COLZAG[k];
			m0 |= 1 << v;
			blk[v] = s;
			++k;
			++nc;
		} else
			break;
	}
	if (!dontskip && blkdiff((short*)sblk, cache) < thresh_) {
		/* skip this block */
		for (;;) {
			if (s != 0) {
				k += r;
				/* Symbol-2 */
				SKIP_BITS(s, nbb, bb);
				++k;
			} else {
				if (r != 15)
				/* end of block */
					break;
				k += 16;
			}
			if (k >= 64)
				break;

			/* Symbol-1 */
			register int v;
			HUFF_DECODE(ht, nbb, bb, v);
			s = v & 15;
			r = v >> 4;
		}
		nbb_ = nbb;
		bb_ = bb;

		return (0);
	}
		
#if NCC != 6
@BUG@
#endif
	((u_int*)cache)[0] = sblk[0];
	((u_int*)cache)[1] = sblk[1];
	((u_int*)cache)[2] = sblk[2];

	for (;;) {
		register int v;
		if (s != 0) {
			k += r;
			/* Symbol-2 */
			GET_BITS(s, nbb, bb, v);
			s = huff_EXTEND(v, s);
			v = COLZAG[k];
			blk[v] = s;
			++nc;

#ifdef INT_64
			m0 |= (INT_64)1 << v;
#else
			if (v < 32)
				m0 |= 1 << v;
			else
				m1 |= 1 << (v - 32);
#endif
			++k;
		} else {
			if (r != 15)
				/* end of block */
				break;

			k += 16;
		}
		if (k >= 64)
			break;

		/* Symbol-1 */
		HUFF_DECODE(ht, nbb, bb, v);
		s = v & 15;
		r = v >> 4;
	}
#ifdef INT_64
	*mask = m0;
#else
	mask[0] = m0;
	mask[1] = m1;
#endif
	nbb_ = nbb;
	bb_ = bb;

	return (nc);
}

/*
 * Skip over a block.  Don't even update the dc predictor.
 */
int JpegDecoder::huffskip(component& p)
{
	register int bb = bb_;
	register int nbb = nbb_;
	u_short* ht = dcht_[p.dc_tbl_no];
	register int s;
	HUFF_DECODE(ht, nbb, bb, s);
	if (s != 0) {
		SKIP_BITS(s, nbb, bb);
	}
	ht = acht_[p.ac_tbl_no];
	for (register int k = 1; k < 64; ) {
		/* Symbol-1 */
		register int v;
		HUFF_DECODE(ht, nbb, bb, v);
		s = v & 15;
		register int r = v >> 4;
		if (s != 0) {
			k += r;
			/* Symbol-2 */
			SKIP_BITS(s, nbb, bb);
			++k;
		} else {
			if (r != 15)
				/* end of block */
				break;
			k += 16;
		}
	}
	nbb_ = nbb;
	bb_ = bb;

	return (0);
}

void JpegDecoder::huffreset()
{
	nbb_ = 0;
	comp_[0].dc = 0;
	comp_[1].dc = 0;
	comp_[2].dc = 0;
}

/*
 * Build a 64k lookup table from the jpeg huffman table described
 * by the arguments.  The table is indexed by the next 16-bits
 * of the input stream.  The entry in the table tells us the
 * length of the next code and its decoded value.
 */
u_short* JpegDecoder::huffbuild(const u_char* bits, const u_char* vals) const
{
	/* Figure C.1: make table of Huffman code length for each symbol */
	/* Note that this is in code-length order. */

	int nsym = 0;
	int huffsize[257];
	for (int codelen = 1; codelen <= 16; ++codelen) {
		for (int i = 1; i <= bits[codelen]; ++i)
			/*
			 * XXX should sanity check that nsym stays
			 * below 256.
			 */
			huffsize[nsym++] = codelen;
	}
	huffsize[nsym] = 0;

	/* Figure C.2: generate the codes themselves */
	/* Note that this is in code-length order. */
  
	int code = 0;
	int si = huffsize[0];
	u_short huffcode[256];
	int p = 0;
	while (p < nsym) {
		while (huffsize[p] == si)
			huffcode[p++] = code++;

		code <<= 1;
		++si;
	}
	/*
	 * Build the direct-map lookup table.
	 */
	u_short *ht = new u_short[65536];
	memset((char*)ht, 0, 65536 * sizeof(u_short));
	for (int sym = 0; sym < nsym; ++sym) {
		int codelen = huffsize[sym];
		int nbit = 16 - codelen;
		int code = huffcode[sym] << nbit;
		int map = (codelen << 8) | vals[sym];
		/*
		 * The low nbit bits are don't cares.
		 * Spin through all possible combos.
		 */
		for (int n = 1 << nbit; --n >= 0; )
			ht[code | n] = map;
	}
	return (ht);
}
