/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/renderer.C,v 1.2 1995/08/16 00:46:32 olson Exp (LBL)";
#endif

#include <stdlib.h>
#include "renderer.h"
#include "color-models.h"

Renderer::Renderer(VideoWindow* w, int inw, int inh)
	: inw_(inw), inh_(inh)
{
	init(w);
}

Renderer::~Renderer()
{
}

static inline int
distance(int a, int b)
{
	a -= b;
	return (a >= 0 ? a : -a);
}

void Renderer::init(VideoWindow* w)
{
	windows_ = w;
	windows_->next_ = 0;
	width_ = w->width();
	height_ = w->height();
	resize(inw_, inh_);
}
	
void Renderer::resize(int w, int h)
{
	inw_ = w;
	inh_ = h;

	/*
	 * Choose a good scale factor.  We scale up or down
	 * by factors of two.  Find the closest scale so that
	 * the input geometry matches to the desired window size.
	 * On a mismatch, the image will be centered in the window.
	 * (This code is unnecessary for decoders that use
	 * hardware scaling, like J-Video, or J300; but just
	 * do it anyway.)
	 */
	scale_ = 0;
	int d = distance(width_, inw_);
	int t = distance(width_, inw_ << 1);
	if (t < d) {
		outw_ = inw_ << 1;
		outh_ = inh_ << 1;
		scale_ = -1;
	} else {
		/*XXX this should be a loop */
		t = distance(width_, inw_ >> 1);
		if (t < d) {
			d = t;
			scale_ = 1;
		}
		t = distance(width_, inw_ >> 2);
		if (t < d) {
			d = t;
			scale_ = 2;
		}
		t = distance(width_, inw_ >> 3);
		if (t < d)
			scale_ = 3;

		outw_ = inw_ >> scale_;
		outh_ = inh_ >> scale_;
	}
	imagesize_ = outw_ * outh_;

	for (VideoWindow* vw = windows_; vw != 0; vw = vw->next_) {
		vw->damage();
		vw->redraw();
	}
}

void Renderer::attach(VideoWindow* w)
{
	w->next_ = windows_;
	windows_ = w;
}

void Renderer::detach(VideoWindow* w)
{
	VideoWindow** p = &windows_;
	while (*p != w)
		p = &(*p)->next_;
	*p = (*p)->next_;
}

int Renderer::contains(VideoWindow* w)
{
	for (VideoWindow* p = windows_; p != 0; p = p->next_)
		if (p == w)
			return (1);
	return (0);
}

SoftwareRenderer::SoftwareRenderer(VideoWindow* w, int inw, int inh, int c,
				   int decimation)
				   
	: Renderer(w, inw, inh), decimation_(decimation)
{
	init(w);
	colorizer_ = 0;
	color(c);
}

SoftwareRenderer::~SoftwareRenderer()
{
	delete image_;
	delete colorizer_;
}

void SoftwareRenderer::init(VideoWindow* w)
{
	Renderer::init(w);
	Tk_Window tk = w->tkwin();
	image_ = new SharedVideoImage(tk, outw_, outh_, imagesize_);
}

void SoftwareRenderer::color(int c)
{
	delete colorizer_;
	ColorModel& cm = ColorModel::instance();
	if (c) {
		if (decimation_ == 422)
			colorizer_ = cm.allocate_422(inw_, inh_, scale_);
		else
			colorizer_ = cm.allocate_411(inw_, inh_, scale_);
	} else
		colorizer_ = cm.gray(inw_, inh_, scale_);
}

void SoftwareRenderer::resize(int w, int h, int c)
{
	int outw = outw_;
	int outh = outh_;
	Renderer::resize(w, h);
	/*
	 * Doing a resize can change the output window size
	 * (but it's not likely because this size is mainly
	 * dependent on the size of the window).
	 */
	if (outw != outw_ || outh != outh_) {
		delete image_;
		Tk_Window tk = windows_->tkwin();
		image_ = new SharedVideoImage(tk, outw_, outh_, imagesize_);
		/*
		 * Make sure the image we just deleted is not
		 * cached in any of our current windows.
		 */
		for (VideoWindow* w = windows_; w != 0; w = w->next_)
			w->setimage(0);
	}
	/*
	 * Make sure a new colorizer is allocated.
	 */
	color(c);
}

void SoftwareRenderer::render(int miny, int maxy) const
{
	if (scale_ >= 0) {
		miny >>= scale_;
		maxy >>= scale_;
	} else {
		miny <<= -scale_;
		maxy <<= -scale_;
	}
	for (VideoWindow* w = windows_; w != 0; w = w->next_)
		w->render(image_, miny, maxy);
}
