/*
 * Copyright (c) 1993 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/switcher.C,v 1.3 1995/08/26 21:16:01 olson Exp (LBL)";

#include <stdlib.h>
#include "inet.h"
#include "video.h"
#include "group-ipc.h"
#include "source-vic.h"
#include "decoder.h"
#include "switcher.h"
#include "Tcl.h"

UserWindow::UserWindow(vicSource* s) : source_(s), touch_(0)
{
	char path[80];
	Tcl& tcl = Tcl::instance();
	tcl.evalf("instantiate %s", s->tcl_method());
	strcpy(path, tcl.result());
	tcl.evalf("winfo toplevel %s", path);
	strcpy(toplev_, tcl.result());
	window_ = VideoWindow::name_to_window(path);
	window_->ra(this);
	window_->ta(this);
	mode_ = USERWIN_MODE_LOCKED;
	interval_ = atoi(Tcl::instance().attr("switchInterval"));
}

UserWindow::~UserWindow()
{
	delete window_;
}

void UserWindow::destroy()
{
	Tcl::instance().evalf("destroy %s", toplev_);
}

void UserWindow::timeout()
{
	switchto(source_->vnext_);
	ssched(interval_);
}

/*
 * called from video widget command handler
 */
void UserWindow::command(VideoWindow* w, int argc, char** argv)
{
	if (argc == 3 && strcmp(argv[1], "key") == 0) {
		if (mode_ != USERWIN_MODE_BROWSE)
			return;
		if (strcmp(argv[2], "comma") == 0 ||
		    strcmp(argv[2], "less") == 0) {
			switchto(source_->vprev_);
		} else if (strcmp(argv[2], "period") == 0 ||
		    strcmp(argv[2], "Return") == 0 ||
		    strcmp(argv[2], "space") == 0 ||
		    strcmp(argv[2], "greater") == 0) {
			switchto(source_->vnext_);
		} else if (strcmp(argv[2], "minus") == 0 ||
			   strcmp(argv[2], "underscore") == 0) {
			if (interval_ > 1)
				--interval_;
		} else if (strcmp(argv[2], "plus") == 0 ||
			   strcmp(argv[2], "equal") == 0) {
			++interval_;
		}
		return;
	}
	if (argc != 2)
		abort();

	if (strcmp(argv[1], "source") == 0) {
		/*
		 * Return the method name for the source.
		 */
		Tcl& tcl = Tcl::instance();
		char* p = tcl.buffer();
		strcpy(p, source_->tcl_method());
		tcl.result(p);
		return;
	}

	int newmode;
	int c = argv[1][0];
	if (c == 'l')
		newmode = USERWIN_MODE_LOCKED;
	else if (c == 'v')
		newmode = USERWIN_MODE_SWITCHED;
	else 
		newmode = USERWIN_MODE_BROWSE;

	if (newmode == mode_)
		return;

	if (mode_ == USERWIN_MODE_SWITCHED)
		Switcher::instance().detach(this);
	if (mode_ == USERWIN_MODE_BROWSE)
		cancel();
	if (newmode == USERWIN_MODE_SWITCHED)
		Switcher::instance().attach(this);
	if (newmode == USERWIN_MODE_BROWSE)
		ssched(interval_);

	mode_ = newmode;
}

void UserWindow::resize(VideoWindow* w, int width, int height)
{
	/*
	 * Delete and insert the window in the decoder, so that
	 * the resize takes effect.
	 */
	Decoder* d = source_->decoder();
	d->detach(w);
	d->attach(w, 0);
}

void UserWindow::name(const char* s)
{
	Tcl::instance().evalf("windowname %s \"%s\"", toplev_, s);
}

void UserWindow::switchto(vicSource* s)
{
	if (source_ == s)
		return;

	source_->detach(this);
	source_ = s;
	source_->attach(this);
}

int Switcher::channel_;
n_long Switcher::local_;
Switcher* Switcher::instance_;

Switcher::Switcher()
	: IPCHandler(1 << GIPC_TALK),
 	  nwin_(0), ipc_(0), clock_(0)
{
}

Switcher::~Switcher()
{
	delete ipc_;
}

Switcher& Switcher::instance()
{
	if (instance_ == 0)
		instance_ = new Switcher;
	return (*instance_);
}

void Switcher::detach(vicSource* s)
{
	int i = 0;
	/*
	 * This is O(n^2) (and could fairly easily be
	 * made linear) but there should never be
	 * many windows in the switcher. 
	 */
	while (i < nwin_) {
		if (windows_[i]->source() == s)
			detach(i);
		else
			++i;
	}
}

void Switcher::detach(UserWindow* p)
{
	if (nwin_ <= 0)
		return;

	int i;
	for (i = 0; i < nwin_; ++i)
		if (windows_[i] == p)
			break;
	detach(i);
}

void Switcher::detach(int i)
{
	--nwin_;
	if (nwin_ == 0) {
		delete ipc_;
		ipc_ = 0;
	}
	for (; i < nwin_; ++i)
		windows_[i] = windows_[i + 1];
}

/*
 * Attach a window to the switcher.  Assume it is not
 * already attached.
 */
void Switcher::attach(UserWindow* p)
{
	if (nwin_ >= MAX_SWITCHED_WINDOWS)
		/*XXX*/
		abort();
	if (nwin_ == 0) {
		ipc_ = new GroupIPC(p->window()->tkwin(), channel_);
		ipc_->attach(this);
	}
	windows_[nwin_++] = p;
}

UserWindow* Switcher::lookupLRU(n_long addr)
{
	int k = 0;
	int v = windows_[0]->touch();
	for (int i = 0; i < nwin_; ++i) {
		if (windows_[i]->source()->addr() == addr) {
			windows_[i]->touch(clock_);
			return (0);
		}
		if (windows_[i]->touch() < v) {
			v = windows_[i]->touch();
			k = i;
		}
	}
	return (windows_[k]);
}

void Switcher::ipc_input(int type, int frompid, u_char* msg, int cc)
{
	if (nwin_ <= 0 || cc < 4)
		return;

	n_long addr = *(n_long*)msg;
	if (addr == local_)
		return;

	/*XXX*/
	vicSource* s = vicSourceManager::instance().consultXXX(addr);
#ifdef notdef
	u_long srcid = 0;
	Source* s = SourceTable::instance().consult(srcid);
#endif
	if (s == 0)
		return;

	++clock_;
	UserWindow* w = lookupLRU(addr);
	if (w == 0)
		return;
	w->switchto(s);
	w->touch(clock_);
}
